// PATCHROM.CPP  Applies a .PAT patch file to an .NES ROM file

#include <stdio.h>
#include <conio.h>

long int filesize(FILE *);

int main(int argc, char *argv[])
{
	FILE *inputrom, *inputpatch, *outputrom; // pointers to the input .NES rom and patch
	char *infname, *patchname, *outfname;
	long int romsize, patchpos, patchsize, startloc, datalength, power,
		   c, c2, maxlength, targetsize=0;
	unsigned char onechar;

	// dynamically allocate 3 char arrays to store filenames
	infname = new char [13];
	patchname = new char [13];
	outfname = new char [13];

	if((argc<3) || (argc>4)) // if invalid number of parameters passed in
	{
		printf("\nPATCHROM - Applies a *.PAT patch file to a given ROM file.\n");
		printf("Usage of the PATCHROM command:  PATCHROM <input ROM> <patch> <output ROM>\n\n");
		printf("If no output filename is given, original input ROM will be overwritten.\n");
		printf("Default ROM extension is *.NES, default patch extension is *.PAT.\n");
		return(0);
	}

	// First read in the source rom filename that was passed in
	for(c=0;c<9;c++)
	{
		infname[c]=argv[1][c];
		if((infname[c]=='.') || (infname[c]=='\0')) // if terminator reached
			break;
	}

	if(c==9) // if source filename too long
	{
		printf("Invalid source filename.  Aborting.\n");
		return(0);
	}

	if(infname[c] == '.') // if there is an extension
	{
		c++;  // :)    move past '.' to get to extension

		for(c2=0;c2<4;c2++)
		{
			infname[c+c2]=argv[1][c+c2];
			if(infname[c+c2]=='\0') // if terminator reached
				break;
		}

		if(c2==4) // if source extension too long
		{
			printf("Invalid source extension.  Aborting.\n");
			return(0);
		}
	}
	else // else add the default .NES extension
	{
		infname[c]   = '.';
		infname[c+1] = 'N';
		infname[c+2] = 'E';
		infname[c+3] = 'S';
		infname[c+4] = '\0';
	}

	// Now read	in the source patch filename that was passed in
	for(c=0;c<9;c++)
	{
		patchname[c]=argv[2][c];
		if((patchname[c]=='.') || (patchname[c]=='\0')) // if terminator reached
			break;
	}

	if(c==9) // if source patch filename too long
	{
		printf("Invalid patch filename.  Aborting.\n");
		return(0);
	}

	if(patchname[c] == '.') // if there is an extension
	{
		c++;  // :)    move past '.' to get to extension

		for(c2=0;c2<4;c2++)
		{
			patchname[c+c2]=argv[2][c+c2];
			if(patchname[c+c2]=='\0') // if terminator reached
				break;
		}

		if(c2==4) // if source patch extension too long
		{
			printf("Invalid patch extension.  Aborting.\n");
			return(0);
		}
	}
	else // else add the default .PAT extension
	{
		patchname[c]   = '.';
		patchname[c+1] = 'P';
		patchname[c+2] = 'A';
		patchname[c+3] = 'T';
		patchname[c+4] = '\0';
	}

	if((inputrom = fopen(infname, "rb")) == NULL) // if source file not present
	{
		printf("Source file not found.  Aborting.\n");
		return(0);
	}

	if((inputpatch = fopen(patchname, "rb")) == NULL) // if source file not present
	{
		printf("Patch file not found.  Aborting.\n");
		return(0);
	}

	romsize = filesize(inputrom); // read size of source file
	patchsize = filesize(inputpatch); // read size of patch file
	fseek(inputpatch, 0, SEEK_SET); // position patch at beginning of file

	// Read in the target size for the file to be patched
	power=0x01; // initialize power multiplier to 1
	for(c=0;c<4;c++) // read in start location
	{
		targetsize += power * fgetc(inputpatch);
		power *= 0x100;
	}

	if(romsize != targetsize) // if they are not the same
	{
		printf("Source ROM size differs from the patch record.  Aborting.\n");
		printf("(Perhaps it is a different version of the ROM?)\n");
		return(0);
	}

	if(argc == 4) // If destination filename given
	{
		// Read in the destination filename passed in
		for(c=0;c<9;c++)
		{
			outfname[c]=argv[3][c];
			if((outfname[c]=='.') || (outfname[c]=='\0')) // if terminator reached
				break;
		}

		if(c==9) // if dest filename too long
		{
			printf("Invalid destination filename.  Aborting.\n");
			return(0);
		}

		if(outfname[c] == '.') // if there is an extension
		{
			c++;  // :)    move past '.' to get to extension

			for(c2=0;c2<4;c2++)
			{
				outfname[c+c2]=argv[3][c+c2];
				if(outfname[c+c2]=='\0') // if terminator reached
					break;
			}

			if(c2==4) // if dest extension too long
			{
				printf("Invalid destination extension.  Aborting.\n");
				return(0);
			}
		}
		else // else add the default .NES extension
		{
			outfname[c]   = '.';
			outfname[c+1] = 'N';
			outfname[c+2] = 'E';
			outfname[c+3] = 'S';
			outfname[c+4] = '\0';
		}

		outputrom = fopen(outfname, "w+b"); // open new outfile with write enabled (if one exists, it will erase it)

		fseek(inputrom, 0, SEEK_SET); // Position source and dest
		fseek(outputrom, 0, SEEK_SET); // at beginning of file

		for(c=0;c<romsize;c++) // copy source to dest file
		{
			onechar = fgetc(inputrom);
			fputc(onechar, outputrom);
		}

		fclose(inputrom); // once copied, dont need original source
	}
	else // else overwrite the input rom
	{
		for(c=0;c<12;c++)
		{
			outfname[c] = infname[c]; // Copy input filename to output filename
			if(outfname[c]=='\0')
				break;
		}

		fclose(inputrom); // close input file before reopening with write enable
		outputrom = fopen(infname, "r+b"); // open source file as output file with write enabled
	}

	// At this point, no longer need the input filenames so deallocate them
	delete[] infname;
	delete[] patchname;

	// NOW, we can start reading the patch data and patching the ROM.
	// (easy) :)

	patchpos = ftell(inputpatch); // find out where we are in the patch file, should be at byte 4

	while((patchsize - patchpos) >= 6) // while 6 byte start+length code remains
	{
		startloc = datalength = 0L; // initialize start & length to 0

		power=0x01; // initialize power multiplier to 1
		for(c=0;c<4;c++) // read in start location
		{
			startloc += power * fgetc(inputpatch);
			power *= 0x100;
		}

		power=0x01; // initialize power multiplier to 1
		for(c=0;c<2;c++) // read in data length
		{
			datalength += power * fgetc(inputpatch);
			power *= 0x100;
		}

		patchpos = ftell(inputpatch); // find out where we are in the patch file

		// check if not enough data remaining or if data would exceed input ROM file boundary
		if((datalength > (patchsize - patchpos)) || ((startloc + datalength) > romsize))
		{
			printf("An error was encountered in the patch file.  Terminating.\n");
			delete[] outfname;
			return(-1); // return with error
		}

		// no errors, so patch this segment.

		fseek(outputrom, startloc, SEEK_SET); // seek to starting location in ROM

		while(datalength--) // loop through the data length
		{
			onechar = fgetc(inputpatch); // read one char from the patch file
			fputc(onechar, outputrom); // write it to the ROM
		}

		patchpos = ftell(inputpatch); // find out where we are in the patch file
	}

	fclose(outputrom);
	fclose(inputpatch);
	printf("%s ROM file patched successfully!\n", outfname);
	delete[] outfname;

	return(0);
}


long int filesize(FILE *stream)
{
	// finds and returns the file size of a file pointed to by stream
	long int curpos, length;

	curpos = ftell(stream);
	fseek(stream, 0L, SEEK_END);
	length = ftell(stream);
	fseek(stream, curpos, SEEK_SET);
	return length;
}

